module CheckProgram(initialProgInfo, processProgram, typeCheck, sigListToSig) where

import Data.List(nub)
import Control.Monad(foldM)
import ErrorMonad
import Term
import Program



distinct :: Eq a => [a] -> Bool
distinct xs = length (nub xs) == length xs


type Vars = Var -> CheckS Type

type Sig = FSym -> CheckS FSymInfo

typeCheck :: Sig -> Vars -> Term -> CheckS Type
typeCheck sig vars (Var x) = vars x
typeCheck sig vars t@(Fun f ts) = do
  (tysIn,tyOut) <- sig f
  tysTs <- mapM (typeCheck sig vars) ts
  assert (tysTs == tysIn) (show t ++ " ill-typed") 
  return tyOut

inferType :: Sig -> Type -> Term -> CheckS [(Var,Type)]
inferType sig ty (Var x) = return [(x,ty)]
inferType sig ty t@(Fun f ts) = do
  (tysIn,tyOut) <- sig f
  assert (length tysIn == length ts) "lengths don't match"
  assert (tyOut == ty) "problem with expected type"
  varsL <- mapM (uncurry (inferType sig)) (zip tysIn ts)
  let vars = nub (concat varsL) 
  assert (distinct (map fst vars)) "conflicting types of variables"
  return vars
  
typeCheckEqn sigma (Var x, r) = failure "var as lhs"
typeCheckEqn sigma (l@(Fun f _), r) = do
  (_,ty) <- sigma f
  vars <- inferType sigma ty l
  tyR <- typeCheck 
     sigma 
     (\ x -> maybeToEither 
         (x ++ " is unknown variable") 
         (lookup x vars))
     r
  assert (ty == tyR) "types of lhs and rhs don't match"
  


processDataDefinition :: ProgInfo -> DataDefinition -> CheckS ProgInfo
processDataDefinition pi@(ProgInfo tys cons defs eqs) (Data ty newCs) = do
  -- check that type is fresh
  assert (not (elem ty tys)) "type is not new"
  let newTys = ty : tys
  -- check distinctness of new constructor names
  assert (distinct (map fst newCs)) "constructors not distinct"
  -- check fresh constructor names
  assert (all (\ (c,_) -> all (/= c) (map fst (cons ++ defs))) newCs) "constructors not new"
  assert (all (\ (_,(tysIn,tyOut)) -> tyOut == ty  
                 && all (\ ty -> elem ty newTys) tysIn) newCs) 
     "problems in types of constructors"
  -- check existence of non-recursive constructor
  assert (any (\ (_,(tysIn,_)) -> all (/= ty) tysIn) newCs)
    "no non-recursive constructor"
  return (ProgInfo newTys (newCs ++ cons) defs eqs)

processDataDefinitions :: ProgInfo -> [DataDefinition] -> CheckS ProgInfo
processDataDefinitions = foldM processDataDefinition

initialProgInfo = ProgInfo [] [] [] []


-- Exercise 2.1

linear :: Term -> Bool
linear t = undefined 

sigListToSig :: SigList -> Sig
sigListToSig slist f = maybeToEither ("unknown symbol: " ++ f) $ lookup f slist


-- Exercise 2.2

checkEquation :: 
  SigList ->       -- defined symbols, including f
  SigList ->       -- constructors
  FSym ->           -- f
  FSymInfo ->      -- type of f
  (Term, Term) ->   -- equation (l,r)
  CheckS ()
checkEquation defs cons f (tysIn, tyOut) (l,r) = undefined

-- Exercise 2.3

processFunctionDefinition :: ProgInfo -> FunctionDefinition -> CheckS ProgInfo
processFunctionDefinition 
  (ProgInfo tys cons defs eqs) 
  (Function f fTy newEqs) = undefined 
  
processFunctionDefinitions :: ProgInfo -> [FunctionDefinition] -> CheckS ProgInfo
processFunctionDefinitions = foldM processFunctionDefinition

processProgram :: FunctionalProg -> CheckS ProgInfo
processProgram (dataDefs, funDefs) = do
  pi <- processDataDefinitions initialProgInfo dataDefs
  processFunctionDefinitions pi funDefs

