module Main(main) where

import FP_Parser
import ErrorMonad
import Term
import Program
import MatchingAndEvaluation
import CheckProgram
import System.IO
import Control.Monad
import Data.Char(isSpace)
import Control.Exception(try, IOException)


trim = dropWhile isSpace

typeCheckGroundTerm (ProgInfo _ cons defs _) t = 
  typeCheck (sigListToSig (defs ++ cons)) (\ _ -> Left "no vars allowed") t

main = mainLoop initialProgInfo


mainLoop pi = do
  putStr "miniGHCi> "
  hFlush stdout
  rline <- getLine
  let line = trim rline
  if take 2 line == ":i" 
   then putStrLn (show pi) >> mainLoop pi
   else if take 2 line == ":q" 
    then putStrLn "Bye"
    else if take 2 line == ":l" 
     then do 
        let fname = trim $ drop 2 line
        srcOrErr <- try $ readFile fname
        case srcOrErr of
          Left (e :: IOException) -> print e >> mainLoop pi
          Right src ->  case parseProg fname src of
            Left e -> putStrLn "error parsing FP-file: " >> putStrLn e >> mainLoop pi
            Right fp -> case processProgram fp of
              Left e -> putStrLn "error with syntactic checks" >> putStrLn e >> mainLoop pi
              Right npi -> putStrLn "loaded program" >> mainLoop npi
     else do
        case parseTerm line of 
          Left e -> putStrLn "cannot parse input term" >> putStrLn e >> mainLoop pi
          Right t -> case typeCheckGroundTerm pi t of
            Left e -> putStrLn "type error in input term" >> putStrLn e >> mainLoop pi
            Right _ -> do
              putStrLn $ show $ evaluate (equations pi) t
              mainLoop pi
